from cscmiko.tools.ssh import SSHManager
from cscmiko.tools.config import check_config_result
from cscmiko.features.base_component import FeatureSet


class CiscoDevice(object):
    """
    base class for cisco Device managers,
    """
    device_type = None

    def __init__(self, host, username, password):
        self.host = host
        connection_dict = {
            "device_type": self.device_type,
            "ip": host,
            "username": username,
            "password": password
        }
        self.conn = SSHManager(connection_dict)

    def disconnect(self):
        self.conn.disconnect()

    def get_command_output(self, command):
        """
        get output of command from device ,
        :param command:
        :return: list of dict or false if connection failed .
        """
        return self.conn.get_command(command)

    def send_commands_config(self, cmds, save):
        """
        send config commands to device
        :param cmds: list: commands strings
        :return: str: execution output
        """
        output = self.conn.send_commands_list(cmds)
        if save:
            self.conn.save_config()
        return output

    def _get_all_cmds(self):
        """try:
        return all cmds added to all self.FeatureSet , and return them
        :return: list of all added commands on that device
        """
        results = []
        for name, obj in vars(self).items():
            if isinstance(obj, FeatureSet):
                results += obj.cmds
        return results

    def commit(self, save=False):
        """
        commit config changes to device,
        :return: (is_ok,err_msg): is_ok is bool() , err_msg is str()
        """
        # get all config changes and set them
        all_cmds = self._get_all_cmds()
        if not all_cmds:
            return False, "No changes to commit"
        # execute the commands
        output = self.send_commands_config(all_cmds, save)
        if not output:
            return False, "Connection to device failed"
        return check_config_result(output)

    def save(self):
        self.conn.save_config()

    def reboot(self):
        """
        reboot the device
        :return:
        """
        pass

    def ping(self, ip):
        """
        ping an ip address from the device
        :param ip:
        :return: bool
        """
        out = self.get_command_output(f"ping {ip}")
        if 'Success rate is 0' in out:
            return False
        return True

    def traceroute(self, ip):
        pass

    def backup(self, file):
        """
        take backup from device and stor it in file location
        :param file:
        :return:
        """
        pass

    def restore(self, file):
        """
        retore config on device from the given fie
        :param file:
        :return:
        """
        pass

    def get_sysname(self):
        """
        get host name
        :return:
        """
        pass

    def get_version(self):
        pass

    def get_sysuptime(self):
        pass



"""
Config component are the features in cisco devices like (routes,acl,vlans ... etc)
"""
from cscmiko.tools.config import render_command
from functools import wraps


def validate_cmd_inputs(kwargs):
    # validate add(),delete() and update() inputs are strings
    for kwarg in kwargs:
        assert isinstance(kwarg, str), f"config inputs should be string not {type(kwarg)}"


def _check_configurable(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not args[0].conf_template:
            raise TypeError(f"feature {type(args[0])._feature_name} is not configurable")

        return func(*args, **kwargs)

    return wrapper


class Feature(object):
    """
    base object of single component like a Vlan , a route
    """

    def __init__(self, output):
        """
        set self the attributes we get from the device output dict,
        output example : {'id':'1', 'description':'vlan 1 description', 'interfaces':['ethernet1','ethernet2']}
        then we do
        self.id = 1
        self.description = 'vlan 1 description'
        self.interfaces = ['ethernet1','ethernet2']
        :param output:
        """
        for key, value in output.items():
            if not value:
                value = None
            setattr(self,key, value)

    @property
    def deserialize(self):
        """
        deserialize the object to dict
        :return:
        """
        return vars(self)

    def __str__(self):
        return str(self.deserialize)

    def __getitem__(self, item):
        return getattr(self, item)

    def __eq__(self, other):
        if isinstance(other, dict):  # __equal__ should accept dict or Feature object
            return self.deserialize == other
        elif isinstance(other, Feature):
            return self.deserialize == other.deserialize
        return False


class FeatureSet(object):
    """
    base object for a list of components like (vlans|interfaces)
    we need this object to group all vlans under one attribute in device manager ,
    my_swicth = CatSwitch(host='4.71.144.98', username='admin', password='J3llyfish1')
    my_swicth.sync_vlans()

    my_switch now will have vlans(which is FeatureSet object) attribute which group all vlan objects ,
    FeatureSet will has it's own methods which applied on all it's children ,
    for example

    my_switch.vlans.count  --> give the count of all vlans
    my_switch.vlans.all --> return list of all vlan objects

    my_switch.vlans.add(id="1",name="aa")

    you can loop through vlans

    for vlan in my_Switch.vlans:
        print(vlan.id)

    """
    model = Feature
    conf_template = ""
    _feature_name = ""

    cmds = []

    @property
    def count(self):
        return len(self)

    def __init__(self, component_dicts):
        self.all = []
        for i in component_dicts:
            self.all.append(self.model(i))

    @_check_configurable
    def add(self, **kwargs):
        """
        add config to device , exampe my_switch.vlans.add(id="1",name="vlan1")
        :param kwargs: config parameters
        :return:
        """
        validate_cmd_inputs(kwargs)
        kwargs.update({"action": "add"})
        cmds = render_command(self.conf_template, kwargs)
        self.cmds += cmds

    @_check_configurable
    def delete(self, **kwargs):
        """
        delete configuration feature , example my_switch.vlans.delete(id="1")
        :param kwargs:
        :return:
        """
        validate_cmd_inputs(kwargs)
        kwargs.update({"action": "delete"})
        cmds = render_command(self.conf_template, kwargs)
        self.cmds += cmds

    @_check_configurable
    def update(self, **kwargs):
        """update device configuration"""
        validate_cmd_inputs(kwargs)
        kwargs.update({"action": "update"})
        cmds = render_command(self.conf_template, kwargs)
        self.cmds += cmds

    def filter(self, **kwargs):
        results = self.all.copy()  # .all should be copied as we will be poping results while iterating through self.all

        for item in self.all:
            for key, value in kwargs.items():
                if not hasattr(self.all[0], key):
                    raise AttributeError(f"{self._feature_name} does not have attribute '{key}' ")
                if item[key].lower() != value.lower():
                    index = results.index(item)
                    results.pop(index)
                    break

        return results

    def __getitem__(self, item):
        return self.all[item]

    def __len__(self):
        return len(self.all)

    def __iter__(self):
        self.i = 0
        return self

    def __next__(self):
        if self.i >= len(self.all):
            raise StopIteration

        obj = self.all[self.i]
        self.i += 1
        return obj



"""
this module has the layer 2 config components ,
it has more business logic on the components like :
vlan.has_interface --> return true of that particular vlan has an interface
interface.is_svi --> return true of that particular interface is an svi interface
"""

from .base_component import FeatureSet, Feature
from cscmiko.features import utils


class Vlan(Feature):
    """single vlan model"""

    def __eq__(self, other):
        return isinstance(other, Vlan) and self.id == other.id

    @property
    def has_interface(self):
        if not self.interfaces:
            return False


class Vlans(FeatureSet):
    """
    device vlans manager class ,
    model attribute is to specify which model this class belong to
    """
    _feature_name = 'vlans'
    model = Vlan

    conf_template = "vlan.j2"


class Interface(Feature):
    """single interface model"""
    name: str
    link_status: str
    protocol_status: str
    hardware_type: str
    address: str
    ip_address: str
    mtu: int
    duplex: str
    speed: str
    bandwidth: int
    input_rate: int
    output_rate: int

    def __init__(self, int_dict):
        super().__init__(int_dict)
        cleand = []
        self.name = utils.translate_interface_name(self.name)
        self.link_status = utils.clean_link_status(self.link_status)
        self.protocol_status = utils.clean_link_status(self.protocol_status)
        self.mtu = int(self.mtu)
        self.bandwidth = int(self.bandwidth.replace(' Kbit', ''))
        self.input_rate = int(self.input_rate)
        self.output_rate = int(self.output_rate)
        self.ip_address, self.netmask = utils.cidr_to_netmask(self.ip_address)

    @property
    def is_svi(self):
        if self.hardware_type == 'EtherSVI':
            return True
        return False

    @property
    def is_port_channel(self):
        if self.hardware_type == 'EtherChannel':
            return True
        return False

    @property
    def is_loopback(self):
        if self.hardware_type == 'Loopback':
            return True
        return False

    @property
    def is_up(self):
        if self.link_status == 'up' and self.protocol_status == 'up':
            return True
        return False


class Interfaces(FeatureSet):
    """multiple interface models class"""
    _feature_name = 'interfaces'
    model = Interface
    conf_template = 'interface.j2'

    def bounce(self, id):
        self.cmds += [f"interface {id}", "shutdown", "no shutdown"]

    @property
    def down_list(self):
        return [i for i in self.all if not i.is_up]

    @property
    def svi_list(self):
        return [i for i in self.all if i.is_svi]

    @property
    def port_channel_list(self):
        return [i for i in self.all if i.is_port_channel]

    @property
    def loopback_list(self):
        return [i for i in self.all if i.is_loopback]


class CdpNrighbor(Feature):
    neighbor: str
    neighbor_ip: str
    platform: str
    remote_port: str
    local_port: str
    software_version: str

    def __init__(self, cdps):
        super().__init__(cdps)
        self.neighbor = utils.extract_device_name(self.neighbor)
        self.remote_port = utils.translate_interface_name(self.remote_port)
        self.local_port = utils.translate_interface_name(self.local_port)
        self.software_version = utils.exteract_software_version(self.software_version)


class CdpNeighbors(FeatureSet):
    _feature_name = 'cdp_neighbors'
    model = CdpNrighbor

    def get_cdps_for_port(self, port):
        return [i for i in self.all if i.local_port == port]

    def get_cdps_for_neighbor(self, neighbor_name):
        return [i for i in self.all if i.neighbor == neighbor_name]

    def get_cdp_by_ip(self, ip):
        return [i for i in self.all if i.neighbor_ip == ip]


class Vtp(Feature):
    pass


class Vpc(Feature):

    @property
    def is_up(self):
        if self.status == 'up':
            return True
        return False


class Vpcs(FeatureSet):
    model = Vpc
    _feature_name = 'vpcs'

    def get_vpc_by_port(self, port):
        return [i for i in self.all if i.port == port]


class Stp(Feature):
    pass


class Stps(FeatureSet):
    model = Stp
    _feature_name = 'spanning_tree'

    def get_by_vlan_id(self, id):
        return [stp for stp in self.all if stp.vlan_id == id]

    def get_by_interface(self, interface):
        for stp in self.all:
            if stp.interface == interface:
                return stp

    def get_by_port_status(self, status):
        """status = FWD|BLK"""
        return [stp for stp in self.all if stp.status == status]

    def get_blocked(self):
        return [stp for stp in self.all if stp.status == 'BLK']

    def get_forwarded(self):
        return [stp for stp in self.all if stp.status == 'FWD']

    def get_designated_ports(self):
        return [stp for stp in self.all if stp.status == 'Desg']

    def get_root_ports(self):
        return [stp for stp in self.all if stp.status == 'Root']



import socket
import struct
import re

def clean_link_status(status):
    if 'down' in status:
        return 'down'
    if 'up' in status:
        return 'up'
    return 'unknown'


def translate_interface_name(name):
    """ Define shortcutting interface name """
    if 'tengigabitethernet' in name.lower():
        name = name.replace('TenGigabitEthernet', 'Te')
        return name

    if 'gigabitethernet' in name.lower():
        name = name.replace('GigabitEthernet', 'Gi')
        return name

    if 'fastethernet' in name.lower():
        name = name.replace('FastEthernet', 'Fa')
        return name

    if 'ethernet' in name.lower():
        name = name.replace('Ethernet', 'Eth')
        return name
    return name


def cidr_to_netmask(cidr):
    if not cidr:
        return None, None
    network, net_bits = cidr.split('/')
    host_bits = 32 - int(net_bits)
    netmask = socket.inet_ntoa(struct.pack('!I', (1 << 32) - (1 << host_bits)))
    return network, netmask

def exteract_software_version(descr):
    version = re.compile('Version ([0-9A-Za-z\(\)\.]*)')
    version = version.search(descr)
    if version:
        return version.group(1)
    return None

def extract_device_name(name):
    if '(' in name or '.' in name:
        match = re.search(r'^([^.(]+?)(\s*[.(])',name)
        if match:
            name = match.group(1)
    return name



# """
# SNMP tools
# """
#
# from easysnmp import Session
# from collections import defaultdict
#
#
# class SNMPManager:
#     """
#     snmp manager , take hostname and community and return dict of results
#     example :
#     snmp_client = SNMPManager('192.168.1.1','private')
#     interfaces = snmp_client.get_fields(name='ifName',mtu='mtu',admin_status='ifAdminStatus')
#     interfaces will equal to ; {'oid_index':'11', 'name':'Ethernet1/0', 'mtu':'1500', 'admin_status':'up'}
#     """
#
#     def __init__(self, hostname, community):
#         self.snmp = Session(hostname=hostname, community=community, version=2, use_sprint_value=True)
#
#     def get_fields(self, **kwargs):
#         """
#         method takes kwargs of field_name and oid name (name='ifName',mtu='mtu',admin_status='ifAdminStatus')
#         :param kwargs: field_name and it's oid name
#         :return: dict: {'oid_index':'11', 'name':'Ethernet1/0', 'mtu':'1500', 'admin_status':'up'}
#         """
#
#         inner_res = defaultdict(dict)
#         for key in kwargs:
#             res = self.snmp.walk(oids=kwargs[key])
#             for item in res:
#
#                 if 'ipAdEntIfIndex' in item.oid:
#                     print(item)
#                     inner_res[item.value].update({key: item.oid_index})
#
#                 else:
#                     inner_res[item.oid_index].update({key: item.value})
#
#         return dict(inner_res)



name = "cscmiko"
import sys
import os
from .devices import *

__version__ = '0.0.9'
# Verify Python Version
try:
    if not sys.version_info.major == 3:
        raise RuntimeError('cscmiko requires Python3')
except AttributeError:
    raise RuntimeError('cscmiko requires Python3')

# set fetch-template env variable
dir_path = os.path.dirname(os.path.realpath(__file__))
os.environ["NET_TEXTFSM"] = dir_path + "/fetch-templates/templates"




"""
this module has the security config components ,
it has more business logic on the components
"""

from .base_component import Feature, FeatureSet
from collections import defaultdict


class Rule(Feature):
    """
    single rule of an access-list
    """
    pass


class ACL:
    """
    an access-list which at least has one rule object
    """
    rules = []
    name = ''


    def __init__(self, name, rules):
        self.name = name
        for rule in rules:
            if not rule['line_num']:  # the root acl should not be counted as rule
                continue
            self.rules.append(Rule(rule))

    def is_extended(self):
        pass


class AccessLists(FeatureSet):
    """
    all device access-lists
    """
    _feature_name = 'access_lists'
    model = ACL
    conf_template = 'acl.j2'

    def __init__(self, component_dicts):
        acl = defaultdict(list)
        for i in component_dicts:
            acl_name = i.pop('name')
            acl[acl_name].append(i)
        for key, value in acl.items():
            self.all.append(ACL(name=key, rules=value))



"""
this module contain the devices managers, for different switch models (Cat,nexus ... etc)
device manager is the user interface to device , where you can fetch config components, push config , backup restore..etc
"""

from .base_device import CiscoDevice
from cscmiko.features import layer2, layer3, security, system
from abc import ABC
from cscmiko.exceptions import CscmikoNotSyncedError, CscmikoInvalidFeatureError

_INVETORY_CMD = "show version"
_VLAN_CMD = "show vlan"
_INTERFACE_CMD = "show interface"
_ROUTE_CMD = "show ip route"
_CDP_CMD = "show cdp neighbors detail"
_BGP_CMD = "show ip bgp"
_OSPF_CMD = "show ip ospf neighbor"
_ACL_CMD = "show ip access-list"
_VRF_CMD = "show vrf"
_VTP_CMD = "show vtp status"
_CPU_CMD = "show processes cpu"
_VPC_CMD = "show vpc"
_MODULE_CMD = "show module"
_STP_CMD = "show spanning-tree"


class _CiscoSwitch(CiscoDevice, ABC):
    """
    Base Cisco Switch manager ,
    this manager handle Cat switch config fetch , config push ,

    my_swicth = CatSwitch(host='4.71.144.98', username='admin', password='J3llyfish1')
    my_swicth.fetch_cpu_status()
    this example fetch CPU status , and set a cpu_status attibute for myswitch object
    """
    features_list = ['invetory', 'interfaces', 'vlans', 'cdp_neighbors', 'routes', 'access_lists', 'vtp_status',
                     'spanning_tree']

    def __getattr__(self, item):
        """
        this is only for raising CiscoSDKNotSyncedError, as the fetch method need to be called before accessing the
        config attribute (e.g. myswitch.vlans )

        for every config compnent(vlans,vrfs,interfaces ... etc) we have a fetch method listed below ,
        :param item: attribute
        :return:
        """
        if item not in self.features_list:
            raise CscmikoInvalidFeatureError(
                f"{item.replace('fetch_','')} is not a valid feature , available features = {self.features_list}")
        if not item.endswith('s'):
            item = item + 's'
        raise CscmikoNotSyncedError(
            f"{item} is not collected  please make sure to call fetch_{item} before, available features : {self.features_list}")

    # Sync Methods
    # TODO : make the add fetch to base class to have a reusable fetch code

    def fetch_inventory(self):
        print(f"Collecting Inventory details from {self.host} ...")
        inventory_dict = self.get_command_output(_INVETORY_CMD)
        if not inventory_dict:
            print("No inventory details collected")
            return None
        self.inventory = system.Inventory(inventory_dict[0])

    # layer 2 fetch methods
    def fetch_interfaces(self):
        print(f"Collecting Interfaces from {self.host} ...")
        interfaces_dicts = self.get_command_output(_INTERFACE_CMD)
        if not interfaces_dicts:
            print("No interfaces collected")
            return None
        self.interfaces = layer2.Interfaces(interfaces_dicts)

    def fetch_vlans(self):
        print(f"Collecting Vlans from {self.host} ...")
        vlans_dicts = self.get_command_output(_VLAN_CMD)
        if not vlans_dicts:
            print("No vlans collected")
            return None
        self.vlans = layer2.Vlans(vlans_dicts)

    def fetch_cdp_neighbors(self):
        print(f"Collecting CDP neighbors from {self.host} ...")
        cdps_dicts = self.get_command_output(_CDP_CMD)
        if not cdps_dicts:
            print("No cdp neighbors collected")
            return None
        self.cdp_neighbors = layer2.CdpNeighbors(cdps_dicts)

    # Layer 3 fetch methods
    def fetch_routes(self):
        print(f"Collecting Routes from {self.host} ...")
        routes_dicts = self.get_command_output(_ROUTE_CMD)
        if not routes_dicts:
            print("No Routes collected")
            return None
        self.routes = layer3.Routes(routes_dicts)

    # security fetch methods
    def fetch_access_lists(self):
        print(f"Collecting access-lists from {self.host} ...")
        acls_dicts = self.get_command_output(_ACL_CMD)
        if not acls_dicts:
            print("No acls collected")
            self.access_lists = None
            return None
        self.access_lists = security.AccessLists(acls_dicts)

    def fetch_spanning_tree(self):
        print(f"Collecting spanning-tree from {self.host} ...")
        stp_dict = self.get_command_output(_STP_CMD)
        if not stp_dict:
            print("No stp collected")
            self.spanning_tree = None
            return None
        self.spanning_tree = layer2.Stps(stp_dict)

    def fetch_vtp_status(self):
        print(f"Collecting vtp status from {self.host} ...")
        vtp_dicts = self.get_command_output(_VTP_CMD)
        if not vtp_dicts:
            print("No vlans collected")
            return None
        self.vtp_status = layer2.Vtp(vtp_dicts[0])

    def fetch(self):
        """
        this call all the fetch_methods incase you want to fetch all components ,
        :return:
        """

        self.fetch_interfaces()
        self.fetch_vlans()
        self.fetch_spanning_tree()
        self.fetch_cdp_neighbors()
        self.fetch_routes()
        self.fetch_access_lists()
        self.fetch_vtp_status()


class CatSwitch(_CiscoSwitch):
    """
    Catalyst Switch device manager which hold it's own fetch methods in addition to base CiscoDevice fetch methods
    """
    device_type = 'cisco_ios'
    features_list = _CiscoSwitch.features_list + ['fetch_cpu_status', 'bgp_neighbors', 'ospf_neighbors', 'vrfs']

    def fetch_cpu_status(self):
        print(f"Collecting cpu status from {self.host} ...")
        cpu_dict = self.get_command_output(_CPU_CMD)
        if not cpu_dict:
            print("No cpu status collected")
            return None
        self.cpu_status = system.Cpu(cpu_dict[0])

    def fetch_bgp_neighbors(self):
        print(f"Collecting BGP neighbors from {self.host} ...")
        bgps_dicts = self.get_command_output(_BGP_CMD)
        if not bgps_dicts:
            print("No BGP collected")
            self.bgp_neighbors = None
            return None
        self.bgp_neighbors = layer3.BgpNeighbors(bgps_dicts)

    def fetch_ospf_neighbors(self):
        print(f"Collecting OSPF neighbors from {self.host} ...")
        ospfs_dicts = self.get_command_output(_OSPF_CMD)
        if not ospfs_dicts:
            print("No OSPF collected")
            self.ospf_neighbors = None
            return None
        self.ospf_neighbors = layer3.OspfNeighbors(ospfs_dicts)

    def fetch_vrfs(self):
        print(f"Collecting VRFs from {self.host} ...")
        vrfs_dicts = self.get_command_output(_VRF_CMD)
        if not vrfs_dicts:
            print("No VRFS collected")
            self.vrfs = None
            return None
        self.vrfs = layer3.Vrfs(vrfs_dicts)

    def fetch(self):
        super().fetch()
        self.fetch_cpu_status()
        self.fetch_ospf_neighbors()
        self.fetch_bgp_neighbors()
        self.fetch_vrfs()


class NexusSwitch(_CiscoSwitch):
    """
    Nexus 9K and 7k Switch device manager which hold it's own fetch methods in addition to base CiscoDevice fetch methods
    """
    device_type = 'cisco_nxos'
    features_list = _CiscoSwitch.features_list + ['modules', 'vpcs']

    def fetch_modules(self):
        print(f"Collecting Modules from {self.host} ...")
        modules_dicts = self.get_command_output(_VRF_CMD)
        if not modules_dicts:
            print("No Modules collected")
            self.modules = None
            return None
        self.modules = system.Modules(modules_dicts)

    def fetch_vpc(self):
        print(f"Collecting vpcs from {self.host} ...")
        vpc_dicts = self.get_command_output(_VPC_CMD)
        if not vpc_dicts:
            print("No vpcs collected")
            self.modules = None
            return None
        self.vpcs = layer2.Vpcs(vpc_dicts)

    def fetch(self):
        super().fetch()
        self.fetch_modules()
        self.fetch_vpc()



"""
this module has the layer 3 config components ,
it has more business logic on the components like :
Route.is_eigrp --> return true of that particular route is eigrp route
BgpNeighbors.suppressed_list --> return list of suppressed bgp objects.
"""

from .base_component import Feature, FeatureSet
from . import utils


class Route(Feature):
    """single route feature"""
    protocol: str
    type: str
    network: str
    netmask: str
    distance: int
    metric: str
    nexthop_ip: str
    nexthop_interface: str
    uptime: str

    def __init__(self, output):
        super().__init__(output)
        self.protocol = utils.ROUTE_PROTOCOL.get(self.protocol, 'unknown')
        self.type = self.type or None
        self.network, self.netmask = utils.cidr_to_netmask(self.network + '/' + self.netmask)
        if self.distance is not None:
            self.distance = int(self.distance)
        if self.metric is not None:
            self.metric = int(self.metric)
        if self.nexthop_interface is not None:
            self.nexthop_interface = utils.translate_interface_name(self.nexthop_interface)

    @property
    def is_connected(self):
        return self.protocol == 'connected'

    @property
    def is_eigrp(self):
        return self.protocol == 'eigrp'

    @property
    def is_ospf(self):
        return self.protocol == 'ospf'

    @property
    def is_static(self):
        return self.protocol == 'static'

    @property
    def is_bgp(self):
        return self.protocol == 'bgp'

    @property
    def is_rip(self):
        return self.protocol == 'rip'


class Routes(FeatureSet):
    """FeatureSet : group of route Features"""
    model = Route
    _feature_name = 'routes'
    conf_template = 'route.j2'

    @property
    def ospf_list(self):
        return [i for i in self.all if i.is_ospf]

    @property
    def eigrp_list(self):
        return [i for i in self.all if i.is_eigrp]

    @property
    def direct_connected_list(self):
        return [i for i in self.all if i.is_connected]

    @property
    def static_list(self):
        return [i for i in self.all if i.is_static]

    @property
    def rip_list(self):
        return [i for i in self.all if i.is_rip]

    def get_routes_of_next_hope_ip(self, ip):
        return [i for i in self.all if ip == i.nexthop_ip]

    def get_routes_of_next_hope_if(self, interface):
        return [i for i in self.all if interface == i.nexthop_if.lower()]

    def get_routes_of_network(self, network):
        return [i for i in self.all if network == i.network]


class Bgp(Feature):
    """Bgp feature"""

    origin:str

    def __init__(self,output):
        super().__init__(output)
        if self.origin is not None:
            if self.is_egp_origin:
                self.origin = 'egp'
            elif self.is_igp_origin:
                self.origin = 'igp'
            elif self.is_incomplete:
                self.origin = 'incomplete'


    @property
    def is_igp_origin(self):
        if self.origin == 'i':
            return True
        return False

    @property
    def is_egp_origin(self):
        if self.origin == 'e':
            return True
        return False

    @property
    def is_incomplete(self):
        if self.origin == '?':
            return True
        return False

    @property
    def is_best(self):
        if self.path_selection == '<':
            return True
        return False

    @property
    def is_valid(self):
        if self.status == '*':
            return True
        return False

    @property
    def is_internal(self):
        if self.route_source == 'i':
            return True
        return False

    @property
    def is_suppressed(self):
        if self.status == 's':
            return True
        return False


class BgpNeighbors(FeatureSet):
    """FeatureSet: set of bgp Features"""
    model = Bgp
    _feature_name = 'bgp_neighbors'

    @property
    def suppressed_list(self):
        return [i for i in self.all if i.is_suppressed]

    @property
    def internal_list(self):
        return [i for i in self.all if i.is_internal]

    @property
    def valid_list(self):
        return [i for i in self.all if i.is_valid]

    @property
    def incomplete_list(self):
        return [i for i in self.all if i.is_incomplete]

    @property
    def best_list(self):
        return [i for i in self.all if i.is_best]

    @property
    def egp_list(self):
        return [i for i in self.all if i.is_egp]

    @property
    def igp_list(self):
        return [i for i in self.all if i.is_igp]

    def get_bgp_by_network(self, network):
        for i in self.all:
            if i.network == network:
                return i


class Ospf(Feature):
    """ospf neighbor Feature"""
    pass


class OspfNeighbors(FeatureSet):
    """FeatureSet: a set of ospf neighbors Feature"""
    model = Ospf
    _feature_name = 'ospf_neighbors'


class Vrf(Feature):
    """Vrf Feature"""
    pass


class Vrfs(FeatureSet):
    """FeatureSet: of vrfs Feature"""
    model = Vrf
    _feature_name = 'vrfs'
